#include "MPU6886.h"
#include <Kalman.h>
#include "FastLED.h"
#include <WiFi.h>
#include <WebServer.h>
#include <Preferences.h>


WebServer server(80);

const char ssid[] = "S-1 DC-sensorless";  // SSID
const char pass[] = "password";   // password

const IPAddress ip(192, 168, 5, 50);      // IPアドレス
const IPAddress subnet(255, 255, 255, 0); // サブネットマスク

#define button 39
#define ImeasPin 22
#define PWM_pin 23
#define PH 33
#define led_pin 27
#define NUM_LEDS 25

CRGB leds[NUM_LEDS];
MPU6886 IMU;

unsigned long oldTime = 0, loopTime, nowTime;
float dt;


//調整パラメータ
//------------------------------
float Kp = 15;
float Kd = 15;
float Kw = 0.8;
float IDRS = 1.0;
int delayTime = 2;
//------------------------------


int pwmDuty;
int GetUP = 0;
int GetUpCnt = 0;
float M;
float Aj = 0.0;


float accX = 0, accY = 0, accZ = 0;
float gyroX = 0, gyroY = 0, gyroZ = 0;
float temp = 0;

float theta_acc = 0.0;
float theta_dot = 0.0;

Kalman kalmanY;
float kalAngleY, kalAngleDotY;

float R = 0.8; //モータ抵抗値[ohm]
float Vi = 0.0;
float Imeas = 0.0;

float rotSpeed, rotEst = 0.0;

Preferences preferences;


//加速度センサから傾きデータ取得 [deg]
float get_theta_acc() {
  IMU.getAccelData(&accX,&accY,&accZ);
  //傾斜角導出 単位はdeg
  theta_acc  = atan(-1.0 * accX / accZ) * 57.29578f;
  return theta_acc;
}

//Y軸 角速度取得
float get_gyro_data() {
  IMU.getGyroData(&gyroX,&gyroY,&gyroZ);
  theta_dot = gyroY;
  return theta_dot;
}


//Core0
void display(void *pvParameters) {
  for (;;){
    server.handleClient();
    
    //LED表示
    for(int i = 0; i < NUM_LEDS; i++){
      leds[i] = CRGB::Black;
    }
      
    if(kalAngleY > 20.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5] = CRGB::Red;
      }
    }else if(kalAngleY <= 20.0 && kalAngleY > 12.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5] = CRGB::Green;
      }
    }else if(kalAngleY <= 12.0 && kalAngleY > 4.0){
      for(int i = 0; i < 5; i++){
       leds[i * 5 + 1] = CRGB::Green;
      }
    }else if(kalAngleY <= 4.0 && kalAngleY > 1.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 2] = CRGB::Green;
      }
    }else if(abs(kalAngleY) <= 1.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 2] = CRGB::Blue;
      }
    }else if(kalAngleY >= -4.0 && kalAngleY < -1.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 2] = CRGB::Green;
      }
    }else if(kalAngleY >= -12.0 && kalAngleY < -4.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 3] = CRGB::Green;
      }
    }else if(kalAngleY >= -20.0 && kalAngleY < -12.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 4] = CRGB::Green;
      }
    }else if(kalAngleY < -20.0){
      for(int i = 0; i < 5; i++){
        leds[i * 5 + 4] = CRGB::Red;
      }
    }
    FastLED.show();
    
    //Serial.println(Imeas);
    
    disableCore0WDT();
  }
}

//ブラウザ表示
void handleRoot() {
  String temp ="<!DOCTYPE html> \n<html lang=\"ja\">";
  temp +="<head>";
  temp +="<meta charset=\"utf-8\">";
  temp +="<title>S-1 DC-sensorless</title>";
  temp +="<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">";
  temp +="<style>";
  temp +=".container{";
  temp +="  max-width: 500px;";
  temp +="  margin: auto;";
  temp +="  text-align: center;";
  temp +="  font-size: 1.2rem;";
  temp +="}";
  temp +="span,.pm{";
  temp +="  display: inline-block;";
  temp +="  border: 1px solid #ccc;";
  temp +="  width: 50px;";
  temp +="  height: 30px;";
  temp +="  vertical-align: middle;";
  temp +="  margin-bottom: 20px;";
  temp +="}";
  temp +="span{";
  temp +="  width: 120px;";
  temp +="}";
  temp +="button{";
  temp +="  width: 100px;";
  temp +="  height: 40px;";
  temp +="  font-weight: bold;";
  temp +="  margin-bottom: 20px;";
  temp +="}";
  temp +="</style>";
  temp +="</head>";
  
  temp +="<body>";
  temp +="<div class=\"container\">";
  temp +="<h3>S-1 DC-sensorless</h3>";
  
  //Kp
  temp +="Kp<br>";
  temp +="<a class=\"pm\" href=\"/KpM\">-</a>";
  temp +="<span>" + String(Kp) + "</span>";
  temp +="<a class=\"pm\" href=\"/KpP\">+</a><br>";

  //Kd
  temp +="Kd<br>";
  temp +="<a class=\"pm\" href=\"/KdM\">-</a>";
  temp +="<span>" + String(Kd) + "</span>";
  temp +="<a class=\"pm\" href=\"/KdP\">+</a><br>";

  //Kw
  temp +="Kw<br>";
  temp +="<a class=\"pm\" href=\"/KwM\">-</a>";
  temp +="<span>" + String(Kw) + "</span>";
  temp +="<a class=\"pm\" href=\"/KwP\">+</a><br>";

  //IDRS
  temp +="IDRS<br>";
  temp +="<a class=\"pm\" href=\"/IDRSm\">-</a>";
  temp +="<span>" + String(IDRS) + "</span>";
  temp +="<a class=\"pm\" href=\"/IDRSp\">+</a><br>";

  //delayTime
  temp +="delayTime<br>";
  temp +="<a class=\"pm\" href=\"/delayTimeM\">-</a>";
  temp +="<span>" + String(delayTime) + "</span>";
  temp +="<a class=\"pm\" href=\"/delayTimeP\">+</a><br>";
  
  temp +="</div>";
  temp +="</body>";
  server.send(200, "text/HTML", temp);
}

void KpM() {
  if(Kp >= 0.1){
    Kp -= 0.1;
    preferences.putFloat("Kp", Kp);
  }
  handleRoot();
}

void KpP() {
  if(Kp <= 30){
    Kp += 0.1;
    preferences.putFloat("Kp", Kp);
  }
  handleRoot();
}

void KdM() {
  if(Kd >= 0.1){
    Kd -= 0.1;
    preferences.putFloat("Kd", Kd);
  }
  handleRoot();
}

void KdP() {
  if(Kd <= 30){
    Kd += 0.1;
    preferences.putFloat("Kd", Kd);
  }
  handleRoot();
}

void KwM() {
  if(Kw >= 0){
    Kw -= 0.1;
    preferences.putFloat("Kw", Kw);
  }
  handleRoot();
}

void KwP() {
  if(Kw <= 30){
    Kw += 0.1;
    preferences.putFloat("Kw", Kw);
  }
  handleRoot();
}

void IDRSm() {
  if(IDRS >= 0.1){
    IDRS -= 0.1;
    preferences.putFloat("IDRS", IDRS);
  }
  handleRoot();
}

void IDRSp() {
  if(IDRS <= 10){
    IDRS += 0.1;
    preferences.putFloat("IDRS", IDRS);
  }
  handleRoot();
}

void delayTimeM() {
  if(delayTime >= 0){
    delayTime -= 1;
    preferences.putInt("delayTime", delayTime);
  }
  handleRoot();
}

void delayTimeP() {
  if(delayTime <= 10){
    delayTime += 1;
    preferences.putInt("delayTime", delayTime);
  }
  handleRoot();
}

void setup() {
  Serial.begin(115200);

  FastLED.addLeds<WS2812B, led_pin, GRB>(leds, NUM_LEDS);
  FastLED.setBrightness(20);
  
  pinMode(button, INPUT);
  pinMode(PH, OUTPUT);

  IMU.Init();

  //フルスケールレンジ
  IMU.SetAccelFsr(IMU.AFS_2G);
  IMU.SetGyroFsr(IMU.GFS_250DPS);

  kalmanY.setAngle(get_theta_acc());

  ledcSetup(0, 20000, 10);
  ledcAttachPin(PWM_pin, 0);
  ledcWrite(0,0);

  preferences.begin("parameter", false);

  //パラメータ初期値取得
  Kp = preferences.getFloat("Kp", Kp);
  Kd = preferences.getFloat("Kd", Kd);
  Kw = preferences.getFloat("Kw", Kw);
  IDRS = preferences.getFloat("IDRS", IDRS);
  delayTime = preferences.getInt("delayTime", delayTime);
 

  WiFi.softAP(ssid, pass);           // SSIDとパスの設定
  delay(100);                        // 追記：このdelayを入れないと失敗する場合がある
  WiFi.softAPConfig(ip, ip, subnet); // IPアドレス、ゲートウェイ、サブネットマスクの設定
  
  IPAddress myIP = WiFi.softAPIP();  // WiFi.softAPIP()でWiFi起動

  server.on("/", handleRoot); 
  
  server.on("/KpP", KpP);
  server.on("/KpM", KpM);
  server.on("/KdP", KdP);
  server.on("/KdM", KdM);
  server.on("/KwP", KwP);
  server.on("/KwM", KwM);
  server.on("/IDRSp", IDRSp);
  server.on("/IDRSm", IDRSm);
  server.on("/delayTimeM", delayTimeM);
  server.on("/delayTimeP", delayTimeP);
  
  server.begin();

  //回転位置検出 タスク
  xTaskCreatePinnedToCore(
    display
    ,  "display"   // A name just for humans
    ,  4096  // This stack size can be checked & adjusted by reading the Stack Highwater
    ,  NULL
    ,  1  // Priority, with 3 (configMAX_PRIORITIES - 1) being the highest, and 0 being the lowest.
    ,  NULL 
    ,  0);
}


void loop() {
  nowTime = micros();
  loopTime = nowTime - oldTime;
  oldTime = nowTime;
  
  dt = (float)loopTime / 1000000.0; //sec

  
  //カルマンフィルタ 姿勢 傾き
  kalAngleY = kalmanY.getAngle(get_theta_acc(), get_gyro_data(), dt);
  
  //カルマンフィルタ 姿勢 角速度
  kalAngleDotY = kalmanY.getRate();

  if (fabs(kalAngleY) < 1.0 && GetUP == 0){
    GetUP = 99;
  }
    
  if(GetUP == 99){
    //ブレーキ
    if(fabs(kalAngleY) > 14.0){
      ledcWrite(0,0);
      Aj = 0.0;
      GetUP = 0;
      rotSpeed = 0.0;
    }

    //回転推定 (フィルタリング)
    rotSpeed += ((Vi - R * Imeas) - rotSpeed) * 0.01;
        
    //モータ回転
    Aj +=  IDRS * rotSpeed / 100.0;
    M = Kp * (kalAngleY + Aj) / 90.0 + Kd * kalAngleDotY / 500.0 + Kw * rotSpeed;
    M = max(-1.0f, min(1.0f, M));
    pwmDuty = 1023 * fabs(M);

    Vi = M * 5.0;
    
    //回転方向
    if(M < 0.0){
      digitalWrite(PH, HIGH);
      ledcWrite(0,pwmDuty);
      Imeas = -1.0 * analogReadMilliVolts(ImeasPin)  / 2.5; //[A]
    }else{
      digitalWrite(PH, LOW);
      ledcWrite(0,pwmDuty);
      Imeas = analogReadMilliVolts(ImeasPin) / 2.5; //[A]
    }
  }else if(GetUP == 0){
    ledcWrite(0,0);
  }
    
  delay(delayTime);
}
